//**********************************************************************
// Copyright 2002 by SoC Solutions, LLC. All rights reserved.          *
// No portion of this material may be reproduced in any form           *
// without the written permission of SoC Solutions.                    *
// All information contained is SoC Solutions company private,         *
// proprietary, and trade secret.                                      *
//**********************************************************************
// Context:     SoC Solutions IP, RTL Verilog for Parallel Port
// Filename:    socParPort.v
// Author:      Chris Stephens
/***********************************************************************
    General Description:
        This module emulates a parallel port.
************************************************************************

    Assumptions :
        The bidirectional data bus is split into two unidirectional
        data buses (parDin and parDout) at the top-levl of the design.
*/

`include "socParPort.h"

module socParPort(
    PRESETn,
    PCLK,
    PADDR,
    PWDATA,
    PWRITE,
    PSEL,
    PENABLE,
    PRDATA,
    blkInt,
    errorN,
    slct,
    pe,
    ackN,
    busyN,
    bidEn,
    intSel,
    slctInN,
    init,
    autoFdXtN,
    strobeN,
    parDin,
    parDout,
    parOeN
    );

    // bus interface
    input  PRESETn;             // system reset, active low
    input  PCLK;                // system clock
    input  [1:0] PADDR;         // lower bits of address bus
    input  [31:0] PWDATA;       // data from processor
    input  PWRITE;              // processor write/not read: 0=rd, 1=wr
    input  PSEL;                // block enable for parallel port
    input  PENABLE;             // wait signal to stretch write cycles
    output [31:0] PRDATA;       // data to processor
    output blkInt;              // interrupt to processor

    input  errorN;              // printer error
    input  slct;                // printer select
    input  pe;                  // printer paper empty
    input  ackN;                // active low acknowledge (of successful data transfer to print buffer)
    input  busyN;               // indicates printer is not ready to accept data

    output strobeN;             // indicates parallel data output valid
    output autoFdXtN;           // auto line feed, when low, printer should feed a line after each printed line
    output init;                // initialize printer
    output slctInN;             // active low printer select

    input  bidEn;               // bidirectional enable
    input  intSel;              // interrupt mode select

    input  [7:0] parDin;        // parallel data in
    output [7:0] parDout;       // parallel data out
    output parOeN;              // parallel output enable

    reg [31:0] PRDATA;
    reg blkIntRaw;              // interrupt output before enable mask

    // writeable parallel port registers
    reg [7:0] parDataReg;
    reg [5:0] control;
    reg [7:0] ioSel;

    // control signals
    wire strobe    = control[0];
    wire autoFdXt  = control[1];
    wire init      = control[2];
    wire slctIn    = control[3];
    wire blkIntEn  = control[4]; // NOTE : blkIntEn does not affect bit [2] of the status register
    wire direction = control[5]; // bits 6 and 7 unimplemented

    wire [7:0] status  = {~busyN,ackN,pe,slct,errorN,~blkIntRaw,1'b1,1'b1};
    wire [7:0] command = {1'b1,1'b1,1'b1,blkIntEn,slctIn,init,autoFdXt,strobe};

    assign strobeN   = ~strobe;
    assign autoFdXtN = ~autoFdXt;
    assign slctInN   = ~slctIn;

    // input/output controls
    assign parDout = parDataReg;
    assign parOeN  = bidEn  ? direction : (ioSel == 8'hAA);


    // register writes
    always @ (posedge PCLK or negedge PRESETn) begin
        if (!PRESETn) begin
            parDataReg <= 8'd0;
            ioSel      <= 8'd0;
            control    <= 6'd0;
        end
        else begin
            if (PSEL & PWRITE & PENABLE) begin
                case (PADDR)
                    `HW_PARALLEL_DATA_OFFSET    : parDataReg <= PWDATA[7:0];
                    `HW_PARALLEL_IOSEL_OFFSET   : ioSel      <= PWDATA[7:0];
                    `HW_PARALLEL_CONTROL_OFFSET : control    <= PWDATA[5:0];
                endcase
            end
        end
    end


    // register reads
    reg readingStatusReg;
    always @ (PSEL or PADDR or parDin or PWRITE or status or command
              ) begin
        if (PSEL) begin
            readingStatusReg = 1'b0;
            case (PADDR)
                `HW_PARALLEL_DATA_OFFSET    :  PRDATA = {24'd0,parDin};
                `HW_PARALLEL_STATUS_OFFSET  :  begin
                                                PRDATA = {24'd0,status};
                                                    if (!PWRITE)
                                                        readingStatusReg = 1'b1;
                                            end
                `HW_PARALLEL_COMMAND_OFFSET : PRDATA = {24'd0,command};
                default                  : PRDATA = 32'd0;
            endcase
        end
        else begin
            PRDATA    = 32'd0;
            readingStatusReg = 1'b0;
        end
    end


    // ackN edge detection
    reg ackNDly1;
    wire fallingEdgeAckN = ~ackN & ackNDly1;
    always @ (posedge PCLK or negedge PRESETn) begin
        if (!PRESETn) begin
            ackNDly1 <= 1'b1;
        end
        else begin
            ackNDly1 <= ackN;
        end
    end


    // logic for determining beginning of status register read
    reg  readingStatusRegDly1;
    wire statusReadStb = ~readingStatusReg & readingStatusRegDly1;
    always @ (posedge PCLK or negedge PRESETn) begin
        if (!PRESETn) begin
            readingStatusRegDly1 <= 1'b0;
        end
        else begin
            readingStatusRegDly1 <= readingStatusReg;
        end
    end


    // interrupt logic
    assign blkInt = blkIntEn ? blkIntRaw : 1'b0;
    always @ (posedge PCLK or negedge PRESETn) begin
        if (!PRESETn) begin
            blkIntRaw <= 1'b0;
        end
        else begin
            if (intSel) begin   // latched mode, latches on falling edge of ackN
                if (fallingEdgeAckN)
                    blkIntRaw <= 1'b1;
                else if (statusReadStb)
                    blkIntRaw <= 1'b0;
            end
            else begin          // ack mode, tracks ackN input
                blkIntRaw <= ~ackN;
            end
        end
    end

endmodule
